{{ sdram-video-20MHz.spin }}
CON
  _clkmode = XTAL1 + PLL16X
  _xinfreq = 5_000_000

  QSIZE = 512
  EMPTY = $FFFF  ' empty queue slot flag

  WAITCYCLES = 16
  WIDTH = 800
  HEIGHT = 480
  GAP = 100      ' pixels per hblank and vblank
OBJ

VAR
  long args [3]

  word queue [QSIZE * 4]   ' hline paint queue
  long qp                  ' insert pointer into queue

  long randstate

PUB Main | col, i, j, k
  init_queue
  randstate := CNT         ' dumb random generator for demo

{ Launch SDRAM cog }
  args[0] := @queue	   ' pass the queue address,
  args[1] := QSIZE         ' and size,
  args[3] := $0            ' and the address in SDRAM of first scanline
  cognew(@entry, @args) 'Launch new cog

  col := $FF               ' colour value
  j := 0
  k := 0

  repeat until j == $0200
    enqueue (0, j, 1024, col)  ' erase one page (use 512 rather than 480 to allow bit mask tests
    j += 1
    if (j & $7F) == 0
      col := (col+25) & $FF    ' change colour every so often

  j := 0
  drain_queue

  ' test pattern with random circles
  repeat
    do_circle (random (500)+150, random (480), random(128)+30, col)
    col := (col+1) &  $FF


{ queue handling code }
PRI init_queue | i
  qp := 0
  repeat i from 0 to QSIZE-1
    queue [i*4+1] := EMPTY

PRI  drain_queue | i
  repeat i from 0 to QSIZE-1
    repeat until queue [i*4+1] == $FFFF


PRI  enqueue (x, y, wid, col)
  repeat until queue [qp*4+1] == $FFFF   ' wait till next slot empty
  queue [qp*4] := col                    ' 4 word fields: colour, x, width, y
  queue [qp*4+2] := x
  queue [qp*4+3] := wid
  queue [qp*4+1] := y & $7FFF            ' the y slot also holds EMPTY flag, so we prevent y == EMPTY
  qp ++
  if qp => QSIZE
    qp -= QSIZE


PRI  sine (i)
  i |= 1
  if i < 0
    return -sine(-i)
  if i => $40000000
    return sine ($80000000-i)
  return word [$E000][(i>>19) & $7FF] ~> 8


PRI  random (n)
  randstate := randstate * 537 + $43853988       ' totally guessed linear congruential parameters(!)
  return (randstate >> 10) // n


PRI  do_circle (x, y, radius, col) | f, dx, dy, x1, y1   ' circle drawing code
  f := 1 - radius
  dx := 1
  dy := -2 * radius
  x1 := 0
  y1 := radius
  circle_slice (x-radius, x+radius, y, col)
  repeat
    if f => 0
      y1 --
      dy += 2
      f += dy
    x1 ++
    dx += 2
    f += dx
    circle_slice (x-x1, x+x1, y+y1, col)
    circle_slice (x-x1, x+x1, y-y1, col)
    circle_slice (x-y1, x+y1, y+x1, col)
    circle_slice (x-y1, x+y1, y-x1, col)
  until x1 => y1
  
PRI  circle_slice (xl, xh, y, col) | w
  w := xh-xl
  if w > 20                  ' the renderer isn't bug free, short spans go wrong!
    enqueue (xl, y, w, col)


DAT

		ORG	0

entry           mov     parm, PAR
                rdlong  queue_beg, parm
                add     parm, #4
                rdlong  queue_end, parm         ' actually the length in entries
                shl     queue_end, #3           ' convert to byte length
                add     queue_end, queue_beg
                add     parm, #4


                mov     OUTA, #0
                or	DIRA, ADDR_MASK
                or      DIRA, CKEMASK           ' ensure CKE driven LOW

                mov	FRQA, CLOCK_FREQ
                mov     PHSA, PHASE
		movs	CTRA, pin_SCLK
		movi	CTRA, #%0_00010_101	' PLL / 4, so 20MHz

                mov	FRQB, CLOCK_FREQ
                mov     PHSB, PHASE
		movs	CTRB, pin_DCLK
		movi	CTRB, #%0_00010_101	' PLL / 1

		mov	VSCL, V5
		mov	VCFG, vconfig_go
		WAITVID	disabled, #0		' force the video hardware to a known state
                or      DIRA, VIDMASK
                or      DIRA, SCLK_MASK		' now clocking in inactive state
                or      DIRA, DCLK_MASK

		mov	temp, CNT
		add	temp, hundred_us
		waitcnt	temp, delay
		call 	#initialize
                mov     starter, #0

top_of_frame
:loop           mov     count, Hheight
                mov     highaddr, starter
                'shr     highaddr, #3
                shl     highaddr, #9
            rdlong highaddr, parm
                and     highaddr, ADDR_MASK
                add     starter, #1
                mov     baaddr, highaddr
                and     baaddr, BAMASK
                mov     lowaddr, baaddr
                mov     VSCL, V5
:scanloop
                WAITVID refresh_comms, refresh_seq  ' synchronize with end of blanking
                call    #burst_read
                WAITVID refresh_comms, refresh_seq
                add     highaddr, #1
                and     highaddr, ADDR_MASK
                mov     baaddr, highaddr
                WAITVID refresh_comms, refresh_seq
                and     baaddr, BAMASK
                mov     lowaddr, baaddr
                WAITVID refresh_comms, refresh_seq
                djnz    count, #:scanloop

                WAITVID refresh_comms, refresh_seq
                mov     cycles_left, GAP_CYCLES
                mov     VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                jmp     #span_interp

BAMASK          long    $00006000
ADDR_MASK       long    $00007FFF

VIDMASK		long	$00FF0000
CKEMASK         long    $00400000
pin_SCLK	long	15
pin_DCLK        long    24
SCLK_MASK	long	$00008000
DCLK_MASK       long    $01000000
vconfig_stop	long    %0_00_1_0_0_000_00000_000000010_011111111
vconfig_go	long    %0_01_1_0_0_000_00000_000000010_011111111



burst_read
                mov     VSCL, V800 ' V1024 ' V4096
                WAITVID read_comms1, read_seq1
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                mov     OUTA, baaddr
                mov     VSCL, V25
                WAITVID read_comms2, read_seq2

                'mov     OUTA, baaddr ' precharge normal
               ' mov     OUTA, #0
burst_read_ret  ret


read_comms1     long    $DC544C5C  ' NOOP, ACTIVE, READ, NOP_DEN
read_seq1       long    %11_11_11_11_11_11_11_11_11_11_11_11_00_10_01_00

read_comms2     long    $DC48D85C  ' NOOP, BT_DEN, PRECHARGE, NOOP_DEN
read_seq2       long    %00_10_00_11_01_11_11




burst_write
                mov     temp, writelen
		add	temp, #4+1
		or	temp, V4096
		mov	VSCL, V5
		mov	OUTA, colour
		WAITVID write_comms0, write_seq0
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                WAITVID write_comms1, write_seq1
		mov	VSCL, temp	' writelen - 5
               mov     OUTA, colour
                WAITVID write_comms1, write_seq1
		mov	VSCL, V25
		WAITVID write_comms2, write_seq2
                mov     OUTA, #0 ' precharge normal
burst_write_ret ret


write_comms0    long    $524C5D5C  ' NOOP, NOP_LCLK, ACTIVE, WRITE_LE
write_seq0      long    %11_10_01

write_comms1    long    $52585E5C  ' NOOP, NOP_LE, BT, WRITE_LE
write_seq1      long    %01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01
narrow_seq      long    %10_01_01_01

write_comms2    long    $48585C  ' NOOP, BT, PRECHARGE
write_seq2      long    %00_10_01




initialize
        	mov	VSCL, V7
 		WAITVID init_comms, init_seq
		mov	OUTA, precharge_all
                nop
                mov	OUTA, modeword
                mov     OUTA, #0
                mov     VSCL, V25 ' V4096
		WAITVID init_comms, #0
                mov     OUTA, #0
                mov     count, #410
:loop
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                djnz    count, #:loop
initialize_ret	ret


modeword	long	%00_0_00_010_0_111
precharge_all	long	$400		' A10 set


nop_comms       long    $5C5C5C5C
init_comms	long	$40485C ' NOOP,PRECHARGE,LOADREG
init_seq	long	%00_10_00_01_00


refresh_comms   long    $44485C ' NOOP, PRECHARGE, AUTO_REFR
refresh_seq     long    %00_10_00_10_00



span_interp	cmp	cycles_left, MIN_CYCLES  wz,wc
	if_b	jmp	#wait_frame          ' run out of time, we're done

		rdlong	w0, queuep
		cmp	w0, HFFFF0000  wc,wz ' is slot empty, then we're done
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_ae	jmp	#span_interp

		add	queuep, #4
		rdlong	w1, queuep	     ' capture slot's data
		sub	queuep, #4
		wrlong	HFFFF0000, queuep    ' mark slot empty

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		call	#handle_span

		add	queuep, #8
		cmp	queuep, queue_end  wz,wc
	if_ae	mov	queuep, queue_beg    ' wraparound

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		jmp	#span_interp

wait_frame      cmps    cycles_left, #0  wz,wc
        if_be   jmp     #top_of_frame
                cmpsub  cycles_left, H800  wc
        if_c    mov     VSCL, V2048
        if_nc   mov     VSCL, cycles_left
        if_nc   add     VSCL, WAITCYCLES
        if_nc   or      VSCL, V4096
        if_nc   mov     cycles_left, #0
                WAITVID refresh_comms, refresh_seq
                jmp     #wait_frame

handle_span	
		mov	highaddr, w0
		shr	highaddr, #16
                and     highaddr, ADDR_MASK
                mov     baaddr, highaddr
                and     baaddr, BAMASK
		mov	colour, w0
		and     colour, #$FF  ' not needed if 16 bit colour
		mov	lowaddr, w1
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		and	lowaddr, H3FF	' assume 10 bit
                or      lowaddr, baaddr
		shr	w1, #16
		mov	writelen, w1

		cmp	writelen, #4  wz,wc   '???
	if_a	jmp	#widespan
narrowspan
		shl	w1, #1
		sub	w1, #8
		neg	w1, w1		' shift count to adapt
		mov	temp, narrow_seq
		shr	temp, w1
		mov	VSCL, V4
		mov	OUTA, colour
		WAITVID write_comms0, write_seq0  ' start the WRITE
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                WAITVID write_comms1, temp   ' nops LE, poss BT
                mov     OUTA, baaddr ' precharge normal
                mov     VSCL, Vwait
		WAITVID write_comms2, #%10 ' just precharge
                sub     cycles_left, #WAITCYCLES+4+4
                jmp     #handle_span_ret
widespan
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                mov     temp, writelen
                cmp     temp, #8  wz,wc
        if_a    jmp     #fullspan

                sub     w1, #4
		shl	w1, #1
		sub	w1, #8
		neg	w1, w1		' shift count to adapt
		mov	temp, narrow_seq
		shr	temp, w1
		mov	VSCL, V4
		mov	OUTA, colour
		WAITVID write_comms0, write_seq0  ' start the WRITE
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                WAITVID write_comms1, write_seq1   ' nops LE
                WAITVID write_comms1, temp   ' nops LE, poss BT
                mov     OUTA, baaddr ' precharge normal
                mov     VSCL, Vwait
		WAITVID write_comms2, #%10 ' just precharge
                sub     cycles_left, #WAITCYCLES+4+4+4

                jmp     #handle_span_ret

fullspan
		sub	temp, #5
                sub     cycles_left, temp
		mov	VSCL, V5
		mov	OUTA, colour
		WAITVID write_comms0, write_seq0
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
               mov     OUTA, colour
	        WAITVID write_comms1, write_seq1
		or	temp, V4096
		mov	VSCL, temp	' writelen - 5
                WAITVID write_comms1, write_seq1
               	mov	VSCL, Vwait
		WAITVID write_comms2, write_seq2
                mov     OUTA, baaddr ' precharge normal
                'mov     OUTA, #0
                sub     cycles_left, #WAITCYCLES+4+4
handle_span_ret ret


hundred_us	long	8000
delay		long	800

CLOCK_FREQ      long    $10000000  ' 5MHz * 16 / 4 = 20MHz PLL
PHASE           long    $00000000

' dataval		long	$000000FF ' 8 bit test pattern, may be 16bit later
writelen        long    250

DEN		long	%10000000 ' $80
LE		long	%00000010 ' $02
LCLK		long	%00000001 ' $01

CEPIN           long    %01000000
CSPIN           long    %00100000
RASPIN          long    %00010000
CASPIN          long    %00001000
WEPIN           long    %00000100

DISABLED	long	%00111100 ' $3C
NOOP		long	%01011100 ' $5C
NOP_DEN		long	%11011100 ' $DC
NOP_LE		long	%01011110 ' $5E
NOP_LCLK	long	%01011101 ' $5D
ACTIVE		long	%01001100 ' $4C
READ		long	%01010100 ' $54
WRITE		long	%01010000 ' $50
WRITE_LE	long	%01010010 ' $52
BT		long	%01011000 ' $58
BT_DEN          long    %11011000 ' $D8
PRECHARGE	long	%01001000 ' $48
AUTO_REFR	long	%01000100 ' $44
LOADREG		long	%01000000 ' $40

Hheight         long    HEIGHT
H3FF            long    $3FF
HFFFF0000       long    $FFFF0000
H800            long    $800
H1FFF		long	$1FFF
H1000           long    $1000
V1              long    $1000 | 1
V4              long    $1000 | 4
V5              long    $1000 | 5
V6              long    $1000 | 6
V7              long    $1000 | 7
V32             long    $1000 | 32
V48             long    $1000 | 48
V25             long    $1000 | 25
Vwait           long    $1000 | WAITCYCLES
V800            long    $1000 | 800
V832            long    $1000 | 832
V1024           long    $1000 | 1024
V2048           long    $1000 | 2048
V4096           long    $1000 | 4096

GAP_CYCLES      long    (WIDTH+GAP)*GAP
MIN_CYCLES      long    WIDTH+125
zero            long    0


starter         res     1

parm            res     1
queuep          res     1
queue_beg       res     1
queue_end       res     1
temp            res     1
count           res     1
ref_count       res     1
w0              res     1
w1              res     1
colour          res     1
cycles_left     res     1
highaddr        res     1
lowaddr         res     1

baaddr          res     1


		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
